/*
 * =================================================================
 * Filename:      cgiirc.c
 * Description:	  CGI:IRC support.
 * Requested by:  LRL
 * Author:        AngryWolf <angrywolf@flashmail.com>
 * =================================================================
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <io.h>
#else
#include <arpa/inet.h>	// inet_pton()
#endif
#include <fcntl.h>
#include "h.h"
#ifdef STRIPBADWORDS
#include "badwords.h"
#endif
#ifdef _WIN32
#include "version.h"
#endif

typedef struct _hostip HostIP;

struct _hostip
{
	HostIP		*prev, *next;
	aClient		*cptr;
	char		*host, *ip;
};

extern void		sendto_one(aClient *, char *, ...);
extern int		find_tkline_match(aClient *, int);

#define CGIIRC_STRING	"CGIIRC"
#define DelHook(x)     	if (x) HookDel(x); x = NULL
#define ircstrdup(x,y)	if (x) MyFree(x); if (!y) x = NULL; else x = strdup(y)

static int		cb_pass(aClient *, char *);
static int		cb_quit(aClient *, char *);
static int		cb_server_quit(aClient *);
static int		cb_unkuser_quit(aClient *, char *);
static int		cb_server_connect(aClient *);
static int		cb_pre_connect(aClient *);
static int		cb_test(ConfigFile *, ConfigEntry *, int, int *);
static int		cb_conf(ConfigFile *, ConfigEntry *, int);
static int		cb_stats(aClient *sptr, char *stats);
static int		cb_rehash();
static inline void	InitConf();
static inline void	FreeConf();
static inline void	free_HostIP(HostIP *p);

static Hook		*HookConfTest, *HookConfRun, *HookStats;
static Hook		*HookConfRehash, *HookPass;
static Hook		*HookPreConnect, *HookServerConnect;
static Hook		*HookUnkUserQuit, *HookQuit, *HookServerQuit;
static HostIP		*HostIPList;
static char		*cgiirc_host;

ModuleHeader MOD_HEADER(cgiirc)
  = {
	"cgiirc",
	"$Id: cgiirc.c,v 1.17 2004/07/12 21:25:03 angrywolf Exp $",
	"CGI:IRC support",
	"3.2-b8-1",
	NULL 
    };

DLLFUNC int MOD_TEST(cgiirc)(ModuleInfo *modinfo)
{
	HookConfTest = HookAddEx(modinfo->handle, HOOKTYPE_CONFIGTEST, cb_test);
	return MOD_SUCCESS;
}

DLLFUNC int MOD_INIT(cgiirc)(ModuleInfo *modinfo)
{
	InitConf();
	HostIPList = NULL;

	HookPass = HookAddEx(modinfo->handle, HOOKTYPE_LOCAL_PASS, cb_pass);
	HookUnkUserQuit = HookAddEx(modinfo->handle, HOOKTYPE_UNKUSER_QUIT, cb_unkuser_quit);
	HookQuit = HookAddEx(modinfo->handle, HOOKTYPE_LOCAL_QUIT, cb_quit);
	HookServerQuit = HookAddEx(modinfo->handle, HOOKTYPE_SERVER_QUIT, cb_server_quit);
	HookPreConnect = HookAddEx(modinfo->handle, HOOKTYPE_PRE_LOCAL_CONNECT, cb_pre_connect);
	HookServerConnect = HookAddEx(modinfo->handle, HOOKTYPE_SERVER_CONNECT, cb_server_connect);
	HookConfRun = HookAddEx(modinfo->handle, HOOKTYPE_CONFIGRUN, cb_conf);
	HookStats = HookAddEx(modinfo->handle, HOOKTYPE_STATS, cb_stats);
	HookConfRehash = HookAddEx(modinfo->handle, HOOKTYPE_REHASH, cb_rehash);

	return MOD_SUCCESS;
}

DLLFUNC int MOD_LOAD(cgiirc)(int module_load)
{
	return MOD_SUCCESS;
}

DLLFUNC int MOD_UNLOAD(cgiirc)(int module_unload)
{
	ListStruct	*next;
	HostIP		*p;

	DelHook(HookConfRehash);
	DelHook(HookStats);
	DelHook(HookConfRun);
	DelHook(HookServerConnect);
	DelHook(HookPreConnect);
	DelHook(HookServerQuit);
	DelHook(HookQuit);
	DelHook(HookUnkUserQuit);
	DelHook(HookPass);
	DelHook(HookConfTest);

	FreeConf();

	for (p = HostIPList; p; p = (HostIP *) next)
	{
		next = (ListStruct *) p->next;
		free_HostIP(p);
	}

	return MOD_SUCCESS;
}

static void InitConf()
{
	cgiirc_host = NULL;
}

static void FreeConf()
{
	if (cgiirc_host)
		MyFree(cgiirc_host);
}

static int cb_rehash()
{
	FreeConf();
	InitConf();
	return 0;
}

static int cb_test(ConfigFile *cf, ConfigEntry *ce, int type, int *errs)
{
	int errors = 0;

	if (type != CONFIG_SET)
		return 0;

	if (!strcmp(ce->ce_varname, "cgiirc-host"))
	{
		if (!ce->ce_vardata)
		{
			config_error("%s:%i: set::cgiirc-host without contents",
					ce->ce_fileptr->cf_filename,
					ce->ce_varlinenum);
			errors++;
		}

		*errs = errors;
		return errors ? -1 : 1;
	}
	else
		return 0;
}

static int cb_conf(ConfigFile *cf, ConfigEntry *ce, int type)
{
	if (type != CONFIG_SET)
		return 0;

	if (!strcmp(ce->ce_varname, "cgiirc-host"))
	{
		if (cgiirc_host)
			MyFree(cgiirc_host);
		cgiirc_host = strdup(ce->ce_vardata);

		return 1;		
	}

	return 0;
}

static int cb_stats(aClient *sptr, char *stats)
{
	if (*stats == 'S')
		sendto_one(sptr, ":%s %i %s :cgiirc-host: %s",
			me.name, RPL_TEXT, sptr->name, cgiirc_host ? cgiirc_host : "<none>");

        return 0;
}

/*
 * Misc functions to handle HostIP structures where
 * the real IP addresses and hostnames of CGI:IRC users
 * are stored.
 */

static HostIP *find_HostIP(aClient *cptr)
{
	HostIP *p;

	for (p = HostIPList; p; p = p->next)
		if (p->cptr == cptr)
			break;

	return p;
}

static inline void free_HostIP(HostIP *p)
{
	DelListItem(p, HostIPList);	
	MyFree(p->host);
	MyFree(p->ip);
	MyFree(p);
}

static void del_HostIP(aClient *cptr)
{
	HostIP *p;

	if ((p = find_HostIP(cptr)))
		free_HostIP(p);
}

/*
 * Here is saved the IP address and hostname provided by
 * a CGI:IRC user (as a password) to a HostIP structure
 * for further examination. That's the only thing I do
 * here, because at first I allow the client to actually
 * connect.
 *
 * NOTE: If a client sends the PASS command, and you
 *       immediately do a /rehash, the HostIP structure
 *       is lost. What's better: not allowing the module
 *       to be unloaded or allowing this rare case to
 *       happen?
 */

static int cb_pass(aClient *sptr, char *password)
{
	if (find_HostIP(sptr))
		return 0;

	if (!strncmp(password, CGIIRC_STRING "_", 7))
	{
		char *ip, *host;

		ip = strchr(password, '_');
		host = (ip && ip[1]) ? strchr(ip + 1, '_') : NULL;

		if (host && host[1])
		{
			HostIP *p = (HostIP *) MyMallocEx(sizeof(HostIP));

			ip++;
			*host++ = 0;

			p->cptr = sptr;
			p->host = strdup(host);
#ifdef INET6
			if (!strchr(ip, ':'))
			{
				p->ip = (char *) MyMalloc(strlen(ip) + 8);
				strcpy(p->ip, "::ffff:");
				strcat(p->ip, ip);
			}
			else
#endif
				p->ip = strdup(ip);

			AddListItem(p, HostIPList);
			MyFree(sptr->passwd);
			sptr->passwd = NULL;

			host[-1] = '_';
			return -1;
		}
	}

	return 0;
}

/*
 * Here I decide if the CGI:IRC client is allowed to connect.
 * Now that I know the new IP and hostname I have to check
 * if they match any allow blocks and bans. Also assign the
 * user to his/her new connection class.
 */

static int cb_pre_connect(aClient *cptr)
{
	HostIP *p;

	if (!(p = find_HostIP(cptr)))
		return 0;

	if (!cgiirc_host || !_match(cgiirc_host, Inet_ia2p(&cptr->ip))
	    || !_match(cgiirc_host, cptr->user->realhost))
	{
		int			i;
		struct hostent		*hp;
		ConfigItem_ban		*bconf;
		ConfigItem_class	*oldclass;

#ifdef INET6
		if (inet_pton(AFINET, p->ip, &cptr->ip) <= 0)
#else
		if ((cptr->ip.S_ADDR = inet_addr(p->ip)) == INADDR_NONE)
#endif
			return exit_client(cptr, cptr, &me,
				"Invalid IP address");

#ifdef PROTO_NICKIP /* Unreal3.2.1 */
		if (cptr->user->ip_str)
			MyFree(cptr->user->ip_str);
		cptr->user->ip_str = strdup(Inet_ia2p(&cptr->ip));
#endif
		cptr->hostp = gethostbyname(p->host);
		strlcpy(cptr->user->realhost, p->host, sizeof cptr->user->realhost);
		strlcpy(cptr->sockhost, p->host, sizeof cptr->sockhost);
		oldclass = cptr->class;

		// *** copied from s_bsd.c and s_user.c
		/*
		 * Verify that the host to ip mapping is correct both ways and that
		 * the ip#(s) for the socket is listed for the host.
		 */
		if ((hp = cptr->hostp))
		{
			for (i = 0; hp->h_addr_list[i]; i++)
				if (!bcmp(hp->h_addr_list[i], (char *)&cptr->ip,
				    sizeof(struct IN_ADDR)))
					break;
			if (!hp->h_addr_list[i])
			{
				sendto_snomask(SNO_JUNK, "IP# Mismatch: %s != %s[%08lx]",
				    Inet_ia2p((struct IN_ADDR *)&cptr->ip), hp->h_name,
				    *((unsigned long *)hp->h_addr));
				hp = NULL;
			}
		}		

		if ((i = AllowClient(cptr, hp, cptr->sockhost, cptr->user->username)))
		{
			if (i == -5)
				return -1;

			sendto_snomask(SNO_CLIENT,
			    "*** Notice -- %s from %s.",
			    i == -3 ? "Too many connections" :
			    "Unauthorized connection", get_client_host(cptr));
			ircstp->is_ref++;
			return exit_client(cptr, cptr, &me,
			    i == -3 ? "This server is full." :
				"You are not authorized to connect to this server");
		}

		/*
		 * Now the client has theoretically been assigned to two classes.
		 * Correct the counter of the "old" class.
		 */
		if (oldclass)
			oldclass->clients--;

		/*
		 * following block for the benefit of time-dependent K:-lines
		 */
#ifdef OFLAG_ADDLINE /* Unreal3.2.1 */
		if ((bconf = Find_ban(cptr, make_user_host(cptr->user->username,
#else
		if ((bconf = Find_ban(make_user_host(cptr->user->username,
#endif
		    cptr->user->realhost), CONF_BAN_USER)))
		{
			ircstp->is_ref++;
			sendto_one(cptr,
			    ":%s %d %s :*** You are not welcome on this server (%s)"
			    " Email %s for more information.",
			    me.name, ERR_YOUREBANNEDCREEP,
			    cptr->name, bconf->reason ? bconf->reason : "",
			    KLINE_ADDRESS);
			return exit_client(cptr, cptr, cptr, "You are banned");
		}

		tkl_check_expire(NULL);
		/* Check G/Z lines before shuns -- kill before quite -- codemastr */
		if ((i = find_tkline_match(cptr, 0)) < 0)
		{
			ircstp->is_ref++;
			return -1;
		}
		find_shun(cptr);
	}

	/*
	 * Connection allowed. Now time to free up the HostIP structure.
	 * Note: if the connection was not allowed, the cb_quit function
	 * would make sure we didn't leave unused structures in the memory.
	 */
	free_HostIP(p);

	return 0;
}

/*
 * The following functions are helpful to avoid memory leaks.
 * Of course del_HostIP() is called two times on a CGI:IRC client,
 * but currently I don't know a far better solution.
 */

static int cb_quit(aClient *sptr, char *comment)
{
	del_HostIP(sptr);
	return 0;
}

static int cb_unkuser_quit(aClient *sptr, char *comment)
{
	del_HostIP(sptr);
	return 0;
}

static int cb_server_quit(aClient *cptr)
{
	del_HostIP(cptr);
	return 0;
}

static int cb_server_connect(aClient *cptr)
{
	del_HostIP(cptr);
	return 0;
}
